#include <pmmintrin.h>  /* SSE3 intrinsics */
#include <stdio.h>      /* printf, srand, rand, RAND_MAX */
#include <stdlib.h>     /* aligned_alloc */
#include <time.h>       /* time, clock_t, clock, CLOCKS_PER_SEC */
#include <math.h>       /* sqrtf */

/* vector size (adjust according to your available memory) */
#define VEC_SZ (1 << 28)

int main(int argc, char *argv[])
{
    clock_t t;

    /* allocate 16-byte aligned buffers of type float (4 bytes) 
     * NOTE: if not aligned, access is slower but use _mm_loadu_ps, 
             _mm_storeu_ps */
    float *x    = aligned_alloc(16, VEC_SZ * sizeof(float));
    float *y    = aligned_alloc(16, VEC_SZ * sizeof(float));
    float *res1 = aligned_alloc(16, VEC_SZ * sizeof(float));
    float *res2 = aligned_alloc(16, VEC_SZ * sizeof(float));

    printf("Buffers allocated: %.1f MB\n", VEC_SZ / 131072.0f);

    /* initialize data buffers with floats in range [0.0, 100.0] */
    srand(time(NULL));
    for (int i=0; i<VEC_SZ; ++i) {
        x[i]    = 100.0f * rand() / RAND_MAX;
        y[i]    = 100.0f * rand() / RAND_MAX;
        res1[i] = 0;
        res2[i] = 0;
    }

    printf("Buffers initialized\n");

    /* start SSE benchamrk */
    t = clock();
    for (int i=0; i<VEC_SZ; i+=4) {
        /* TODO: implement the operations from below using SSE intrinsics
         *       store the result in res1*/
    }
    t = clock() - t;

    printf("SSE operation:    %.3f\n", ((float)t) / CLOCKS_PER_SEC);

    /* start normal benchmark */
    t = clock();
    for (int i=0; i<VEC_SZ; ++i)
        res2[i] = sqrtf(x[i]) / y[i];
    t = clock() - t;

    printf("Normal operation: %.3f\n", ((float)t) / CLOCKS_PER_SEC);

    /* cleanup */
    free(x);
    free(y);
    free(res1);
    free(res2);

    printf("Buffers freed\n");

    return 0;
}